open Lexer
open Lexing
open Typing
open Syntax
open Util
open Preprocessing
open Constraint_solving

let print_position outx lexbuf =
  let pos = lexbuf.lex_curr_p in
  Printf.fprintf outx "%s:%d:%d" pos.pos_fname
    pos.pos_lnum (pos.pos_cnum - pos.pos_bol + 1)

let print_error_position (e:error) : string = 
  match e with 
  | Error (pos1, pos2) -> 
    let s1 = Printf.sprintf "%d:%d" pos1.pos_lnum (pos1.pos_cnum - pos1.pos_bol + 1) in 
    let s2 = Printf.sprintf "%d:%d" pos2.pos_lnum (pos2.pos_cnum - pos2.pos_bol + 1) in 
    pos1.pos_fname ^ ":(" ^ s1 ^ "," ^ s2 ^ ")"
  | ErTODO -> "(POS: TODO)"

let parse_with_error lexbuf =
  try Parser.top Lexer.read lexbuf with
  | SyntaxError msg ->
    Printf.fprintf stderr "%a: %s\n" print_position lexbuf msg;
    None
  | Parser.Error ->
    Printf.fprintf stderr "%a: syntax error\n" print_position lexbuf;
    exit (-1)

let rec parse_and_print lexbuf =
  match parse_with_error lexbuf with
  | Some (ke,p,q) ->

    (try 
      let p_news, p_repl, p_norepl = split_process p in
      let q_news, q_repl, q_norepl = split_process q in
      let old_p = (p_norepl, (p_news,p_repl)) in 
      let old_q = (q_norepl, (q_news,q_repl)) in 
      if uniquenv p && uniquenv q && 
      check_key_env ke &&
      check_repl_scoping p_norepl false  &&
      check_repl_scoping p_news false  &&
      check_repl_scoping p_repl true  &&
      check_repl_scoping q_norepl false  &&
      check_repl_scoping q_news false  &&
      check_repl_scoping q_repl true  then
        
        if scoping ke p
        then
          begin
            let ke = extend_d ke p in 
            let e = extract_e old_p in 
            let e' = extract_e old_q in 
            (*
            print_verbose 2 (process2string (fst p));
            print_verbose 2 ("\n~~~\n");
            print_verbose 2 (process2string (fst q));
            *)
            print_verbose 0 "Typing processes and generating constraints...";
            let css1 = tcp e e' (var_env_empty ()) ke p_norepl q_norepl in 
            let css2 = tcp e e' (var_env_empty ()) ke p_repl q_repl in 
            print_verbose 1 ("\n\nGenerated constraints for non-replicated part : \n\n ");
            print_verbose 1 (constrsetset2string css1);
            print_verbose 1 ("\n\nGenerated constraints for replicated part : \n\n ");
            print_verbose 1 (constrsetset2string css2);
            print_verbose 1 "\n\n";
            if check_consistency e e' ke css1 css2 then
              print_verbose 0 ("\nConstraints are consistent.\n\nThe two processes are trace equivalent.");
            parse_and_print lexbuf
          end
        else
          raise (Type_Error ("Bad scoping: some nonces appear in both bound and unbound positions", ErTODO, ErTODO))
      else
        raise (Type_Error ("Nonces and Variables are not unique",ErTODO,ErTODO))
    with 
     | Repl_Error (s,er,er') -> print_verbose 0 ("REPLICATION ERROR:\n"  ^ s ^ "\n" ^ print_error_position er ^ "\n" ^ print_error_position er'); exit 1
     | Type_Error (s,er,er') -> print_verbose 0 ("TYPE ERROR:\n"  ^ s ^ "\n" ^ print_error_position er ^ "\n" ^ print_error_position er'); exit 1
     | FormatError er -> print_verbose 0 ("FORMAT ERROR:\n" ^ print_error_position er); exit 1
     | Constraint_Error s -> print_verbose 0 ("CONSTRAINT ERROR:\n"  ^ s ); exit 1)
  | None -> ()

let loop filename =
  let inx = open_in filename in
  let lexbuf = Lexing.from_channel inx in
  lexbuf.lex_curr_p <- { lexbuf.lex_curr_p with pos_fname = filename };
  parse_and_print lexbuf;
  close_in inx

let syntax_msg = "Syntax: "^Sys.argv.(0)^" [-v0 | -v1] file.pi"

let main =
  if Array.length Sys.argv = 1 ||
       ((Sys.argv.(1) = "-v0" || Sys.argv.(1) = "-v1" || Sys.argv.(1) = "-v2") && (Array.length Sys.argv = 2)) ||
         (Array.length Sys.argv = 3 && Sys.argv.(1) <> "-v0" && Sys.argv.(1) <> "-v1" && Sys.argv.(1) <> "-v2" && Sys.argv.(1) <> "-v3")|| (Array.length Sys.argv > 3)then
    print_endline syntax_msg
  else        
    let filename = if Array.length Sys.argv = 2 then Sys.argv.(1) else Sys.argv.(2) in
    if Array.length Sys.argv = 3 then
      (if Sys.argv.(1) = "-v0" then
         __verbose_level := 0;
       if Sys.argv.(1) = "-v1" then
         __verbose_level := 1;
       if Sys.argv.(1) = "-v2" then
         __verbose_level := 2;
       if Sys.argv.(1) = "-v3" then
         __verbose_level := 3);
    loop filename
