open Syntax 
open Util


(* Collect all defining occurences of nonces and variables of a process *)
let collectnv (p:process) : ((nonce * label) list * (key * (ty * bool)) list * var list) =
  let rec collectnv_aux (p:process) (r:bool) : ((nonce * label) list * (key * (ty * bool)) list * var list) = match p with 
    | PZero _  -> ([],[],[])
    | PNew (_, n, l, p') -> let (ns,ks,vs) = collectnv_aux p' r in ((n,l)::ns,ks,vs)
    | PNewKey (_, k, t, p') -> let (ns,ks,vs) = collectnv_aux p' r in (ns,(k,(t,r))::ks,vs)
    | POut (_,_,p') -> collectnv_aux p' r
    | PIn (_,x, p') -> let (ns,ks,vs) = collectnv_aux p' r in (ns,ks,x::vs)
    | PPar (_,p1,p2) -> 
        let (ns,ks,vs) = collectnv_aux p1 r in 
        let (ns',ks',vs') = collectnv_aux p2 r in 
        (ns @ ns',ks @ ks', vs @ vs')
  | PRepl (_,p') ->  collectnv_aux p' true
    | PLet (_,x,_,p1,p2) -> 
        let (ns,ks,vs) = collectnv_aux p1 r in 
        let (ns',ks',vs') = collectnv_aux p2 r in 
        (ns @ ns',ks @ ks', x::vs @ vs')
    | PIf (_,_,_,p1,p2) -> 
        let (ns,ks,vs) = collectnv_aux p1 r in 
        let (ns',ks',vs') = collectnv_aux p2 r in 
        (ns @ ns',ks @ ks', vs @ vs')
    | PMatch (_,_,y,p1,z,p2) ->
        let (ns,ks,vs) = collectnv_aux p1 r in 
        let (ns',ks',vs') = collectnv_aux p2 r in 
        (ns @ ns',ks @ ks', y :: z :: vs @ vs') in
  collectnv_aux p false

(* Check if all nonces and variables are unique *)
let uniquenv (p:process) : bool = 
  let ns, ks, vs = collectnv p in
  let (ns', ks', vs') = (List.sort_uniq Pervasives.compare (List.map fst ns), 
                         List.sort_uniq Pervasives.compare (List.map fst ks), 
                         List.sort_uniq Pervasives.compare vs) in 
  List.length ns = List.length ns' && 
  List.length ks = List.length ks' && 
  List.length vs = List.length vs'

(* extract the nonce environment from the process *)
let extract_e (p1, (pn, p2) :process * (process * process)) : nonce_env = 
  let fst3 x = let (x1, x2, x3) = x in x1 in
  let nl = fst3 (collectnv p1) @ fst3 (collectnv p2) @ fst3 (collectnv pn) in
  List.fold_left (fun f (n,l) -> extend f n l) env_empty nl

(* extract the key environment from the process *)
let extend_d (d:key_env) (p:process) : key_env = 
  let _, kl, _= collectnv p in
  List.fold_left (fun f (k,(t,r)) -> extend_key_env f (k,k) t r) d kl

let rec check_key_env (d:key_env) : bool  = 
  let l = KeyEnv.bindings d in 
  let ll_same = 
    List.fold_left 
      (fun b ((k1,k2),(t,r)) -> b && (match t with 
                                  | TKey (LL, _ ) -> k1 = k2 
                                  | _ -> true)) 
      true l in 
  let keys = List.sort_uniq (String.compare) (List.fold_left (fun ks ((k1,k2),l) -> k1::k2::ks) [] l) in 
  let k_same_label = 
    List.fold_left 
      (fun b k -> 
        b && fst (List.fold_left 
          (fun (b', ol) ((k1,k2),(t,r)) -> 
            if k1 = k || k2 = k then 
              (match t with 
              | TKey (l, _) 
              | TSessKey (l, _) 
              | TEqKey (l, _) -> 
                  (match ol with 
                  | Some l' -> l=l', ol
                  | None -> b', Some l)
              | _ -> failwith "Impossible in check_key_env")
            else
              b', ol)
          (true,None)
          l))
      true
      keys in 
  ll_same && k_same_label


(* Check that replication scoping is respected *)

exception Repl_Error of string * error * error

let rec check_repl_scoping (p:process) (repl:bool) : bool = 
  match p with 
  | PZero e  -> true
  | PNew (e, n, l, q) -> 
    (match n with 
    | Nonce _ -> 
      if repl then 
        raise (Repl_Error ("Cannot bind non-star nonce under replication", e, e))
      else
        check_repl_scoping q repl
    | NonceStar n -> 
        if not repl then 
          raise (Repl_Error ("Cannot bind star-nonce without replication", e, e))
        else 
          check_repl_scoping q repl)
  | PNewKey (e, k, t, q) -> check_repl_scoping q repl
  | POut (e, t, q) -> 
      if not repl && not (no_star t) then 
        raise (Repl_Error ("Cannot usd star-terms without replication", e, e))
      else
        check_repl_scoping q repl
  | PIn (e, x, q) -> check_repl_scoping q repl
  | PPar (e, q1, q2) -> 
      check_repl_scoping q1 repl && check_repl_scoping q2 repl
  | PRepl (e, p') -> 
      if repl then 
        raise (Repl_Error ("Illegal nested replication", e, e))
      else
        check_repl_scoping p' true
  | PLet (e, x, d, q1, q2) -> check_repl_scoping q1 repl && check_repl_scoping q2 repl
  | PIf (e, u, v, q1, q2) -> 
      if not repl &&  not (no_star u && no_star v) then
        raise (Repl_Error ("Cannot usd star-terms without replication", e, e))
      else
        check_repl_scoping q1 repl && check_repl_scoping q2 repl

  | PMatch (e, x, y, q1, z, q2) -> check_repl_scoping q1 repl && check_repl_scoping q2 repl


(* returns the list of nonces that appear in p but are not bound in p *)
let unbound_nonces (p:process) : (nonce list * key list) =
(* aux returns the unbound nonces in p which are not in nn *)
  let rec unbound_nonces_aux (p:process) (nn:nonce list) (kk:key list) : (nonce list * key list) =
    match p with
    | PZero _ -> [],[]
    | PNew (_, n, _, p') -> unbound_nonces_aux p' (n :: nn) kk
    | PNewKey (_, k, _, p') -> unbound_nonces_aux p' nn (k :: kk)
    | POut (_, t, p') -> let (nt,kt,_) = (collectnv_t t) in
                         let nt' = List.filter (fun n -> not (List.mem n nn)) nt in
                         let kt' = List.filter (fun k -> not (List.mem k kk)) kt in
                         let nn', kk' = unbound_nonces_aux p' nn kk in
                         nt' @ nn', kt' @ kk'
    | PRepl (_, p') 
    | PIn (_, _, p') -> unbound_nonces_aux p' nn kk
    | PLet (_, _, d, p', p'')  ->
        (match d with 
        | DSDec (_, _, t)
        | DADec (_, _, t)
        | DVer  (_, _, t) -> 
            let (nt,kt,_) = (collectnv_t t) in
             let nt' = List.filter (fun n -> not (List.mem n nn)) nt in
             let kt' = List.filter (fun k -> not (List.mem k kk)) kt in
             let nn', kk' = unbound_nonces_aux p' nn kk in
             let nn'', kk'' = unbound_nonces_aux p'' nn kk in
             nt' @ nn' @nn'', kt' @ kk' @ kk''
        | _ -> let nn1, kk1 = (unbound_nonces_aux p' nn kk) in 
               let nn2, kk2 = (unbound_nonces_aux p'' nn kk) in 
               nn1 @ nn2, kk1 @ kk2)
    | PMatch (_, _, _, p', _, p'') 
    | PPar (_, p', p'') -> let nn1, kk1 = (unbound_nonces_aux p' nn kk) in 
                           let nn2, kk2 = (unbound_nonces_aux p'' nn kk) in 
                           nn1 @ nn2, kk1 @ kk2
    | PIf (_, t, t', p', p'') ->
       let (nt1,kt1,_) = (collectnv_t t) in
       let (nt2,kt2,_) = (collectnv_t t') in
       let (nt,kt) = (nt1 @ nt2), (kt1 @ kt2) in 
       let nt' = List.filter (fun n -> not (List.mem n nn)) nt in
       let kt' = List.filter (fun k -> not (List.mem k kk)) kt in
       let nn1, kk1 = (unbound_nonces_aux p' nn kk) in 
       let nn2, kk2 = (unbound_nonces_aux p'' nn kk) in      
       nt' @ nn1 @ nn2, kt' @ kk1 @ kk2 
  in
  unbound_nonces_aux p [] [] 

(*return true iff the scoping is correct in p, ie a nonce that appears unbound is never declared new *)
let scoping (ke:key_env) (p:process) : bool =
  let unbound = unbound_nonces p in
  let bn, bk, _ = collectnv p in
  let bound_n = List.map fst bn in
  let bound_k = List.map fst bk in
  List.for_all (fun n -> not (List.mem n bound_n)) (fst unbound) &&
  List.for_all (fun k -> not (List.mem k bound_k)) (snd unbound) 


exception FormatError of error

let rec process_replication (p:process) : bool = match p with 
  | PZero e  -> false
  | PNew (e, _, _, p') 
  | PNewKey (e, _, _, p')
  | POut (e, _, p') 
  | PIn (e, _, p') -> process_replication p'
  | PPar (e, p', p'')
  | PLet (e, _, _, p', p'')
  | PIf (e, _, _, p', p'')
  | PMatch (e, _, _, p', _ , p'') -> process_replication p' || process_replication p''
  | PRepl (e, p') -> true


let rec split_process (p:process) : process * process * process = 
  match p with
  | PZero e  -> (PZero e, PZero e, PZero e)
  | PNew (e, n, l, p') -> 
      let p1, p2, p3 = split_process p' in 
      PNew (e, n, l, p1), p2 ,p3
  | PNewKey (e, k, t, p') -> 
      let p1, p2, p3 = split_process p' in 
      PNewKey (e, k, t, p1), p2 ,p3
  | POut (e, _,_) 
  | PIn (e, _,_)
  | PLet (e, _, _, _, _) 
  | PIf (e, _, _, _, _)
  | PMatch (e, _, _, _, _, _) -> 
      if not (process_replication p) then
        (PZero ErTODO, PZero ErTODO, p)
      else
        raise (FormatError e)
  | PPar (e, p', p'') -> 
      if not (process_replication p'') then
        match p' with 
        | PRepl (e, p''') -> (PZero ErTODO, p''', p'')
        | _ -> (PZero ErTODO, PZero ErTODO, p)
      else
        raise (FormatError e)
  | PRepl (e, p') -> 
      if not (process_replication p') then 
        (PZero ErTODO, p', PZero ErTODO)
      else
        raise (FormatError e)
