open Syntax
open Util


(* reduce destructor applications *)
let eval (d:dest) (t:term) : term option = 
  match d with 
  | DSDec (_,x, kt) -> 
    (match t with 
    | VSEnc (_,t', kt') -> 
      if same_term kt kt' then Some t' else None
    | _ -> None)
  | DADec (_,x, kt) -> 
    (match t with 
    | VAEnc (_,t', VPKey (_, kt')) -> 
      if same_term kt kt' then Some t' else None
    | _ -> None)
  | DProj1 (_,p) -> 
    (match t with 
    | VPair (_,t1, t2) -> Some t1
    | _ -> None)
  | DProj2 (_,p) -> 
    (match t with 
    | VPair (_,t1, t2) -> Some t2
    | _ -> None)
  | DVer (_,x, VVerKey (_, kt)) -> 
    (match t with 
    | VSign (_,t', kt') -> 
      if same_term kt kt' then Some t' else None
    | _ -> None)
  | _ -> None


let is_key (t:term) : bool =
  match t with
  | VKey _ -> true
  | _ -> false
       
let is_pubkey (t:term) : bool =
  match t with
  | VPKey (_,t') ->
     (match t' with
      | VKey _ -> true
      | _ -> false)
  | _ -> false

let is_verkey (t:term) : bool =
  match t with
  | VVerKey (_,t') ->
     (match t' with
      | VKey _ -> true
      | _ -> false)
  | _ -> false
  
let rec eval_term_not_bot (t:term) : bool =
  match t with
  | VNonce _ -> true
  | VConst _ -> true
  | VVar _ -> true
  | VKey _ -> true
  | VPKey (_,t') -> is_key t'
  | VVerKey (_,t') -> is_key t'
  | VPair (_,t',t'') -> (eval_term_not_bot t') && (eval_term_not_bot t'')
  | VInl _ -> failwith "tagged unions are not supported"
  | VInr _ -> failwith "tagged unions are not supported"
  | VHash (_,t') -> eval_term_not_bot t'
  | VSEnc (_,t',t'') -> (eval_term_not_bot t') && (is_key t'')
  | VAEnc (_,t',t'') -> (eval_term_not_bot t') && (is_pubkey t'')
  | VSign (_,t',t'') -> (eval_term_not_bot t') && (is_verkey t'')
                      
(* Exception thrown when encountering a typing error *)
exception Type_Error of string * error * error

let rec subtyping (t1:ty) (t2:ty) = 
  print_verbose 3 ("Subtyping :" ^ type2string t1 ^ ", " ^ type2string t2);
  (* SRefl *)
  same_type t1 t2 ||
  match t1, t2 with 
  (* SHigh *)
  | _ , TLabel LH -> true
  (* SPair *)
  | TPair (t1', t1''), TPair (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SPairL *)
  | TPair (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SPairS  + SPairS'*)
  | TPair (t1', t1''), TLabel LS -> subtyping t1' (TLabel LS) ||  subtyping t1'' (TLabel LS)
  (* SSum *)
  | TSum (t1', t1''), TSum (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SSumL *)
  | TSum (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SKey *)
  | TKey (l,t), TLabel l' -> l = l'
  (* SEqKey *)
  | TEqKey (l,t), TKey (l',t') -> l = l' && t = t'
  (* SSesKey *)
  | TSessKey (l,t), TEqKey (l',t') -> l = l' && t = t'
  (* SPubKey *)
  | TPubKey t, TLabel LL -> 
      (match t with 
      | TSessKey _ | TEqKey _ -> true
      | _ -> false)
  (* SVerKey *)
  | TVerKey t, TLabel LL -> 
      (match t with 
      | TSessKey _ | TEqKey _ -> true
      | _ -> false)
  (* SEnc *)
  | TSEnc (t,k), TSEnc (t',k') -> same_type k k' &&  subtyping t t'
  (* SAEnc *)
  | TAEnc (t,k), TAEnc (t',k') -> same_type k k' &&  subtyping t t'
  (* Rule STrans is incomporated directly in the other rules *)
  | _ -> false
   
(* Function to get the key type of an arbitrary term (if it is a key) *)
let infer_key_type (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) : (ty option) =
  match u, v with
  (* Either it is a key syntactically *)
  | VKey (er, k), VKey (er', k') -> lookup_key d (k,k') 
  (* Or a variable bound to key type *)
  | VVar (er, x), VVar (er', y) -> 
      if x = y then
        let t = lookup_var g x in 
        match t with 
        | Some (TKey _) 
        | Some (TSessKey _) 
        | Some (TEqKey _) -> 
            t
        | _ -> None
      else
        None
  | _ -> None

let infer_pubkey_type (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) : (ty option) =
  match u, v with
  (* Either it is a pubkey syntactically *)
  | VPKey (er, k), VPKey (er', k') -> 
      (match infer_key_type e e' g d k k' with
      | Some kt -> Some (TPubKey kt)
      | None -> None)
  (* Or a variable bound to a PubKey type *)
  | VVar (er, x), VVar (er', y) -> 
      if x = y then
        let t = lookup_var g x in 
        match t with 
        | Some (TPubKey _) -> t 
        | _ -> None
      else
        None
  | _ -> None

let rec tct_new (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) (t:ty) : (bool * constr_set) =  
  print_verbose 3 ("tct: " ^ term2string u ^ ", " ^ term2string v);
  let b,c = 
    try 
    (

    match u, v, t with 
    (* THigh *)
    | _ ,_ , TLabel LH -> eval_term_not_bot u && eval_term_not_bot v, []
    (* TOr *)
    | u, v, TOr (t1, t2) -> 
      let b1, c1 = tct_new e e' g d u v t1 in
      let b2, c2 = tct_new e e' g d u v t2 in 
      if b1 then
        true, c1
      else
        b2, c2

    (* TNonce + TNonceL *)
    | VNonce (er,n), VNonce (er',m), TLabel l ->
      (match e n, e' m with 
       | Some l1 (*(l1, _)*), Some l2 (*(l2, _)*) -> 
        if l1 = l2 && l1 = l && (l1 <> LL || n = m) then 
          true, []
        else
          raise (Type_Error ("Types of the nonces in the environment don't match. Type of " ^ nonce2string n ^ " is " ^ label2string l1 ^ " ,Type of " ^ nonce2string m ^ " is " ^ label2string l2, er, er'))
      | None, None -> 
        if n = m then 
          if l = LL then
            true, []
          else
            raise (Type_Error ("Unbound nonce is always Low",er,er'))
        else raise (Type_Error ("Nonce not bound on both sides and not equal", er, er'))
      | None, _ -> raise (Type_Error ("Nonce " ^ nonce2string n ^ " is not bound (on the left)", er, er'))
      | _ -> raise (Type_Error ("Nonce " ^ nonce2string m ^ " is not bound (on the right)", er, er')))

    (* TPair *)
    | VPair (er,n, m), VPair (er',n', m'), TPair (t1, t2) -> 
      let b1,c1 = tct_new e e' g d n n' t1 in
      let b2,c2 = tct_new e e' g d m m' t2 in
      b1 && b2, (c1 @ c2) 

    (* TConst *)
    | VConst (er,c), VConst (er',c'), TLabel LL -> 
      if c = c' then   
        true, []
      else
        raise (Type_Error ("Constants " ^ c ^ " and " ^ c' ^ " are not equal", er, er'))

    (* TInl *)
    | VInl (er,n), VInl (er',m), TSum (t1, t2) ->
      tct_new e e' g d n m t1

    (* TInr *)
    | VInr (er,n), VInr (er',m), TSum (t1, t2) -> 
      tct_new e e' g d n m t2

    (* TKey *)
    | VKey (er, k), VKey (er', k'), TSessKey _ 
    | VKey (er, k), VKey (er', k'), TEqKey _
    | VKey (er, k), VKey (er', k'), TKey _ -> 
      (match lookup_key d (k,k') with  
      | Some t' -> same_type t t', []
      | None -> raise (Type_Error ("key " ^ k ^ " is not bound", er, er')))

    (* TPubKey *)
    | VPKey (er, u'), VPKey (er', v'), TPubKey t' -> 
      (match t' with 
      | TKey (l, t'') | TSessKey (l, t'') | TEqKey (l, t'') ->
        (match tct_new e e' g d u' v' t' with 
        | true, [] -> true, [] (* TODO exists... Check t' in KT? *) 
        | false, _ -> raise (Type_Error ("Key term does not have type expected by pk", er, er'))
        | _ , _ -> raise (Type_Error ("Key term may not produce constraints in pk", er, er')))
      | _ -> raise (Type_Error ("pk expects a key type as argument", er, er')))

    (* TPubKeyL *)
    | VPKey (er, VKey (_ ,k)), VPKey (er', VKey (_ ,k')), TLabel LL -> 
      if k = k' then
        if is_in_keys d k then true, []
        else raise (Type_Error ("Public key pk(" ^ k ^") is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TVKey *)
    | VVerKey (er, u'), VVerKey (er', v'), TVerKey t' -> 
      (match t' with 
      | TKey (l, t'') | TSessKey (l, t'') | TEqKey (l, t'') ->
        (match tct_new e e' g d u' v' t' with 
        | true, [] -> true, [] (* TODO exists... *)
        | false, _ -> raise (Type_Error ("Key term does not have type expected by pk", er, er'))
        | _ , _ -> raise (Type_Error ("Key term does may not produce constraints in pk", er, er')))
      | _ -> raise (Type_Error ("pk expects a key type as argument", er, er')))

    (* TVKeyL *)
    | VVerKey (er, VKey (_, k)), VVerKey (er', VKey (_, k')), TLabel LL -> 
      if k = k' then
        if is_in_keys d k then true, []
        else raise (Type_Error ("Verification key vk(" ^ k ^") is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))

    (* TVar *)
    | VVar (er,x), VVar (er',y), t -> 
      (match lookup_var g x, lookup_var g y, t with
        (* Instantiation of variables as a specialized version of TLR' combined with TLRVar *)
        | Some (TRefine (u1,v1)), Some (TRefine (u2,v2)), _ -> 
          if not (is_star_nonce u1 || is_star_nonce u2 || is_star_nonce v1 || is_star_nonce v2) then 
            tct_new e e' g d u1 v2 t
          else
            (* TLR'* *)
            if (is_star_nonce u1 && is_star_nonce u2 && is_star_nonce v1 && is_star_nonce v2 && x = y) then 
              tct_new e e' g d u1 v2 t
            else
             raise (Type_Error ("cannot instantiate ill-formed refinements", er, er'))
        | Some (TSEnc (t1,t1')), Some (TSEnc (t2,t2')), TLabel LL ->
            if x = y then
              (match t1, t1' with 
              | t3, TKey (LS, t3') 
              | t3, TEqKey (LS, t3') 
              | t3, TSessKey (LS, t3') -> subtyping t3 t3', [u,v]
              | TLabel LL, TLabel LL
              | TLabel LL, TKey (LL, _) 
              | TLabel LL, TEqKey (LL, _) 
              | TLabel LL, TSessKey (LL, _) -> true, []
              | _ -> raise (Type_Error ("Variables have illegal SEnc Type", er, er')))
            else
              raise (Type_Error ("variables " ^ x ^ " and " ^ y ^ " are not equal", er, er'))
        | Some (TAEnc (t1,t1')), Some (TAEnc (t2,t2')), TLabel LL ->
            if x = y then
              (match t1, t1' with 
              | t3, TPubKey (TKey (LS, t3'))
              | t3, TPubKey (TEqKey (LS, t3'))
              | t3, TPubKey (TSessKey (LS, t3')) -> subtyping t3 t3', [u,v]
              | TLabel LL, TLabel LL
              | TLabel LL, TPubKey (TKey (LL, _))
              | TLabel LL, TPubKey (TEqKey (LL, _))
              | TLabel LL, TPubKey (TSessKey (LL, _)) -> true, []
              | _ -> raise (Type_Error ("Variables have illegal AEnc Type", er, er')))
            else
              raise (Type_Error ("variables " ^ x ^ " and " ^ y ^ " are not equal", er, er'))
        | _ -> 
        (if x = y then
          match lookup_var g x with 
          | Some (TOr _)  ->  raise (Type_Error ("Variables may not be output with or type", er, er'))
          | Some t' -> subtyping t' t, []
          | None -> raise (Type_Error ("variable " ^ var2string x ^ " is not bound", er, er'))
        else
          raise (Type_Error ("variables " ^ var2string x ^ " and " ^ var2string y ^ " are not equal", er, er'))))

    (* TEnc *)
    | VSEnc (er, m1, m2), VSEnc (er',n1, n2), TSEnc (t1,t2) ->
      let b1, c1 = tct_new e e' g d m1 n1 t1 in 
      let b2, c2 = tct_new e e' g d m2 n2 t2 in 
      (match b1 && b2, c2 with
      | true, [] -> 
        (match t2 with 
        | TLabel LL | TKey _ | TSessKey _ | TEqKey _ -> true, c1
        | _ -> false, [])
      | false , _ -> false, []
      | true, _ ->  raise (Type_Error ("Key term may not produce constraints in TEnc", er, er')))

    (* TEncH + TEncL *)
    | VSEnc (er, m1, m2), VSEnc (er',n1, n2), TLabel LL ->
      let kt = infer_key_type e e' g d m2 n2 in 
      (match kt with 
      | Some (TKey (LS, t))
      | Some (TSessKey (LS, t))
      | Some (TEqKey (LS, t)) -> 
          let b, c = tct_new e e' g d u v (TSEnc (t, option_get kt)) in
          b, (u,v)::c
      | Some (TKey (LL, _))
      | Some (TSessKey (LL, _))
      | Some (TEqKey (LL, _)) -> tct_new e e' g d u v (TSEnc (TLabel LL, option_get kt))
      | None ->  tct_new e e' g d u v (TSEnc (TLabel LL, TLabel LL))
      | _ -> failwith ("Impossible due to invariant on infer_key_type"))

    (* TAEnc *)
    | VAEnc (er, m1, m2), VAEnc (er',n1, n2), TAEnc (t1,t2) ->
      let b1, c1 = tct_new e e' g d m1 n1 t1 in 
      let b2, c2 = tct_new e e' g d m2 n2 t2 in 
      (match b1 && b2, c2 with
      | true, [] -> 
        (match t2 with 
        | TLabel LL -> true, c1
        | TPubKey t3 -> 
            (match t3 with 
            | TKey _ | TSessKey _ | TEqKey _ -> true, c1
            | _ -> raise (Type_Error ("PKey expects a key in  TAEnc", er, er')))
        | _ -> false, [])
      | false , _ -> false, []
      | true, _ ->  raise (Type_Error ("Key term may not produce constraints in TEnc", er, er')))

    (* TAEncH + TAEncL *)
    | VAEnc (er, m1, m2), VAEnc (er',n1, n2), TLabel LL ->
      let kt = infer_pubkey_type e e' g d m2 n2 in 
      (match kt with 
      | Some (TPubKey (TKey (LS, t)))
      | Some (TPubKey (TSessKey (LS, t)))
      | Some (TPubKey (TEqKey (LS, t))) -> 
          let b, c = tct_new e e' g d u v (TAEnc (t, option_get kt)) in
          b, (u,v)::c
      | Some (TPubKey (TEqKey (_, _))) 
      | Some (TPubKey (TSessKey (_, _))) -> tct_new e e' g d u v (TAEnc (TLabel LL, option_get kt))
      | Some (TPubKey (TKey (_, _))) -> 
         raise (Type_Error ("Could not infer equality of keys in TAEncL", term_error u, term_error v))
      | None ->  tct_new e e' g d u v (TAEnc (TLabel LL, TLabel LL))
      | _ -> failwith ("Impossible due to invariant on infer_pubkey_type"))

    (* TSignH + TSignL*)
    | VSign (er, m1, m2), VSign (er',n1, n2), TLabel LL ->
      let kt = infer_key_type e e' g d m2 n2 in 
      (match kt with 
      | Some (TEqKey (LS, t))
      | Some (TSessKey (LS, t)) ->
          let b1, c1 = tct_new e e' g d m1 n1 t in
          let b2, c2 = tct_new e e' g d m1 n1 (TLabel LL) in
          b1 && b2 , (u,v)::(c1 @ c2)
      | _ -> 
          let b1, c1 = tct_new e e' g d m1 n1 (TLabel LL) in
          let b2, c2 = tct_new e e' g d m2 n2 (TLabel LL) in
          if c2 <> [] then 
            raise (Type_Error ("Key in TSignL must produce empty constraints", term_error u, term_error v))
          else
            b1 && b2 , c1)

    (* THash + THashL *)
    | VHash (er,n), VHash (er',m), TLabel LL -> 
        (* Heuristic: we use THash if hash contains an S term directly under pair *)
        if s_under_pair e g d n && s_under_pair e' g d m then
          true, [u,v]
        else
          tct_new e e' g d n m (TLabel LL)

    (* TLR *)
    | u, v, TRefine (tl, tr) -> 
        print_verbose 2 ("Refinment Check! " ^  term2string u ^ ", " ^ term2string v ^ " : " ^ term2string tl ^ ", " ^ term2string tr);
        if ground u && ground v then
          if well_formed_refinement u v then
            same_term u tl && same_term v tr, []
          else
            raise (Type_Error ("Refinement Type is not well formed", term_error u, term_error v))
        else
          raise (Type_Error ("Terms for refinement types must be ground",term_error u, term_error v))



    (***** TSub Rules *****)

    (* SHigh not needed *)
    (* SRefl not needed *)
    (* STrans implicitly by recursion *)
    (* SPair not needed *)

    (* SPairL *)
    | VPair (er,n, m), VPair (er',n', m'), TLabel LL -> 
      tct_new e e' g d u v (TPair (TLabel LL,TLabel LL)) 

    (* SPairS *)
    | VPair (er,n, m), VPair (er',n', m'), TLabel LS -> 
      let b1, c1 = tct_new e e' g d u v (TPair (TLabel LS,TLabel LH)) in 
      if b1 then b1, c1 else
      tct_new e e' g d u v (TPair (TLabel LH,TLabel LS)) 

    (* STU not needed *)

    (* STUL *)
    | VInl (er,n), VInl (er',n'), TLabel LL -> 
      tct_new e e' g d u v (TSum (TLabel LL,TLabel LL)) 
    | VInr (er,n), VInr (er',n'), TLabel LL -> 
      tct_new e e' g d u v (TSum (TLabel LL,TLabel LL)) 

    (* SKey *)
    | VKey (er, k), VKey (er',k'), TLabel l ->
      if k = k' then
        match get_key_label d k with 
        | Some l' -> l = l', []
        | _ -> raise (Type_Error ("key " ^ k ^ " is not bound", er, er'))
      else
        raise (Type_Error ("keys " ^ k ^ " and " ^ k' ^ " are not equal", er, er'))


    | _ -> raise (Type_Error ("Term Typing Failure : " ^ term2string u ^ " ~ " ^ term2string v ^ " /:/ " ^ type2string t, term_error u, term_error v))     
    )
  with
    | Type_Error (s,er,er') -> print_verbose 2 ("WARNING!\n"  ^ s); false, [] 

  in 
  print_verbose 2 ("Typing term : " ^ term2string u ^ ", " ^ term2string v ^ " : " ^ type2string t ^  " returns " ^ string_of_bool b); 
  if c <> [] then 
    print_verbose 2 ("Got non-empty constraints"); 
  b,c




let tcd (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (de:dest) (de':dest): ty =  
  match de,de' with 
  | DSDec (er,x, VKey (_, k)), DSDec (er',x2, VKey (_, k2)) -> 
    if x = x2 then
      (match lookup_key d (k,k2), lookup_var g x with
      (* DDecH *)
      | Some (TKey (LS, t)), Some (TLabel LL) -> failwith "Impossible: This should be covered on process level"
      (* DDecL *)
      | Some (TKey (LL, t)), Some (TLabel LL) 
      | Some (TSessKey (LL, t)), Some (TLabel LL) 
      | Some (TEqKey (LL, t)), Some (TLabel LL) -> 
          if k = k2 then 
            TLabel LL
          else
            raise (Type_Error ("Key differing",er,er'))
      (* DDecT *)
      | Some (TSessKey (l, t)) , Some (TSEnc (t',TSessKey (l', t''))) -> 
          if same_type t t'' && l = l' then
            t'
          else
            raise (Type_Error ("Keys don't match in DDecT",er,er'))
      | Some t, Some t' -> raise (Type_Error ("Symetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Symetric decryption with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | DSDec (er,x, VVar (_, y)), DSDec (er',x2, VVar (_, y2)) -> 
    if x = x2 && y = y2 then
      (match lookup_var g y, lookup_var g x with
      (* DDecH' *)
      | Some (TSessKey (LS, t)), Some (TLabel LL) -> t
      (* DDecL' *)
      | Some (TSessKey (LL, _)), Some (TLabel LL) 
      | Some (TLabel LL), Some (TLabel LL) -> TLabel LL
      (* DDecT *)
      | Some (TSessKey (l, t)) , Some (TSEnc (t',TSessKey (l', t''))) -> 
          if same_type t t'' && l = l' then
            t'
          else
            raise (Type_Error ("Keys don't match in DDecT",er,er'))
      | Some t, Some t' -> raise (Type_Error ("Symetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Symetric decryption with unbound variables",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | DADec (er,x, VKey (_, k)), DADec (er',x2, VKey (_, k2)) -> 
    if x = x2  then
      (match lookup_key d (k,k2), lookup_var g x with
      (*DADecH*)
      | Some (TKey (LS, t)), Some (TLabel LL) -> failwith "Impossible: This should be covered on process level"
      (*DADecL*)
      | Some (TKey (LL, t)), Some (TLabel LL)
      | Some (TSessKey (LL, t)), Some (TLabel LL) 
      | Some (TEqKey (LL, t)), Some (TLabel LL) -> 
        if k = k2 then
          TLabel LL
        else
          raise (Type_Error ("Key is differing",er,er'))
      (* DADecT *)
      | Some (TSessKey (l,t)) , Some (TAEnc (t',TPubKey (TSessKey (l',t'')))) -> 
          if same_type t t'' && l = l' then
            t'
          else
            raise (Type_Error ("Keys don't match in DDecT",er,er'))

      | Some t, Some t' -> raise (Type_Error ("Assymetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Assymetric decryption with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable is differing",er,er'))

  | DADec (er,x, VVar (_, y)), DADec (er',x2, VVar (_, y2)) -> 
    if x = x2  && y = y2 then
      (match lookup_var g y, lookup_var g x with
      (*DADecH'*)
      | Some (TSessKey (LS, t)), Some (TLabel LL) -> TOr (t, TLabel LL)
      (*DADecL'*)
      | Some (TSessKey (LL, _)), Some (TLabel LL) 
      | Some (TLabel LL), Some (TLabel LL) -> TLabel LL
      (* DADecT' *)
      | Some (TSessKey (l,t)) , Some (TAEnc (t',TPubKey (TSessKey (l',t'')))) -> 
          if same_type t t'' && l = l' then
            t'
          else
            raise (Type_Error ("Keys don't match in DDecT",er,er'))

      | Some t, Some t' -> raise (Type_Error ("Assymetric Decryption of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Assymetric decryption with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable is differing",er,er'))

  | DVer (er,x, VVerKey (_, VKey (_, k))), DVer(er',x2,VVerKey (_, VKey (_, k2))) -> 
    if x = x2 && k = k2 then
    (match lookup_key d (k,k), lookup_var g x with
      (* DCheckH *)
      | Some (TSessKey (LS, t)), Some (TLabel LL) 
      | Some (TEqKey (LS, t)), Some (TLabel LL) 
      | Some (TKey (LS, t)), Some (TLabel LL) -> t
      (* DCheckL *)
      | Some (TSessKey (LL, t)), Some (TLabel LL)
      | Some (TEqKey (LL, t)), Some (TLabel LL)
      | Some (TKey (LL, t)), Some (TLabel LL) -> TLabel LL
      (* ELSE *)
      | Some t, Some t' -> raise (Type_Error ("Verification of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Verification with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable or key is differing",er,er'))

  | DVer (er,x, VVar (_,y)), DVer(er',x2,VVar (_,y2)) -> 
    if x = x2 && y = y2 then
    (match lookup_var g y, lookup_var g x with
      (* DCheckH' *)
      | Some (TVerKey (TSessKey (LS, t))), Some (TLabel LL) 
      | Some (TVerKey (TEqKey (LS, t))), Some (TLabel LL) -> t
      (* DCheckL' *)
      | Some (TVerKey (TSessKey (LL, _))), Some (TLabel LL) 
      | Some (TVerKey (TEqKey (LL, _))), Some (TLabel LL) 
      | Some (TLabel LL), Some (TLabel LL) -> TLabel LL
      (* ELSE *)
      | Some t, Some t' -> raise (Type_Error ("Verification of incompatible types. Key type is : " ^ type2string t ^ ". Variable type is : " ^ type2string t',er,er'))
      | _ -> raise (Type_Error ("Verification with unbound key or variable",er,er')))
    else
      raise (Type_Error ("Destructor Variable or key is differing",er,er'))

  | DProj1 (er,x), DProj1 (er',x2) -> 
    if x = x2 then
      (match lookup_var g x with 
      | Some (TPair (t1, t2)) ->  t1
      | Some (TLabel LL) -> TLabel LL
      | _ -> raise (Type_Error ("Projection not applicable",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | DProj2 (er,x), DProj2 (er',x2) -> 
    if x = x2 then
      (match lookup_var g x with 
      | Some (TPair (t1, t2)) -> t2
      | Some (TLabel LL) ->  TLabel LL
      | _ -> raise (Type_Error ("Projection not applicable",er,er')))
    else
      raise (Type_Error ("Destructor Variable differing",er,er'))

  | _ -> raise (Type_Error ("Destructors do not match",ErTODO,ErTODO))


(* collect all variables with or types  that appear at top level of the
 * processes *)
let collect_top_or (g:var_env) (p:process) (q:process) : var list = 
  let vars = 
    let thrd3 (_, _, x) = x in
    match p, q with 
    | POut (_,u, p'), POut (_,v, q') -> 
      let v_u = thrd3 (collectnv_t u) in 
      let v_v = thrd3 (collectnv_t v) in 
      v_u @ v_v
    | PLet (_,x, de, p', p''), PLet (_,x', de', q', q'') ->
      [dest_var de; dest_var de']
    | PIf (_,u, v, p', p''), PIf (_,u', v', q', q'') -> 
      let v_u = thrd3 (collectnv_t u) in 
      let v_u' = thrd3 (collectnv_t u') in 
      let v_v = thrd3 (collectnv_t v) in 
      let v_v' = thrd3 (collectnv_t v') in 
      v_u @ v_u' @ v_v @ v_v'
    | PMatch (_,x, y, p', z, p''), PMatch (_,x', y', q', z', q'') -> 
      [x;x']
    | _ -> []
  in List.sort_uniq Pervasives.compare 
    (List.filter (fun x -> match lookup_var g x with Some TOr _ -> true | _ -> false) vars)


(* This represents a weaker form of instantiation *)
let rec infer_refinement (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (v:term) : ty option = 
  match u,v with 
  | VVar (_,x), VVar (_,y) ->
      (match lookup_var g x, lookup_var g y with
      | Some (TRefine (u1,u2)), Some (TRefine (v1, v2)) -> 
        if not (is_star_nonce u1 || is_star_nonce u2 || is_star_nonce v1 || is_star_nonce v2) || x = y then
          Some (TRefine (u1, v2))
        else 
          None
      | _ -> None)
  | _ -> if  well_formed_refinement u v then 
           Some (TRefine (u, v))
         else
           None



let incompatible_types (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (u:term) (u':term) (v:term) (v':term) : bool = 
  (* PIfS*)
  let b1, _ = tct_new e e' g d u u' (TLabel LL) in
  let b2, _ = tct_new e e' g d v v' (TLabel LS) in
  if b1 && b2 then
    true
  else
    (* PIfS symmetric version *)
    let b1, _ = tct_new e e' g d u u' (TLabel LS) in
    let b2, _ = tct_new e e' g d v v' (TLabel LL) in
    if b1 && b2 then 
      true
    else
      (* PIfI *)
      match (infer_refinement e e' g d v v') with
      | Some (TRefine (v1, v2)) ->
          let pairs2 = match v1, v2 with 
            | VPair _, _ -> true
            | _ , VPair _ -> true
            | _ -> print_verbose 2 "none are pairs"; false in 
          let pairs1 = fst (tct_new e e' g d u u' (TPair (TLabel LH, TLabel LH)))
          in pairs1 && not pairs2
      | _ -> false

(* returns true iff g1 is a subset of g2 *)
exception Not_subset
        
let var_env_subset (g1:var_env) (g2:var_env) : bool =
  try
    let _ =
    VarEnv.merge
      (fun x ot1 ot2 ->
        match ot1,ot2 with
        | Some t1, Some t2 ->
           if same_type t1 t2 then None
           else raise Not_subset
        | Some _, None -> raise Not_subset
        | None, Some _ -> None
        | None, None -> None)
      g1 g2 in
    true
  with
  | Not_subset -> false
                

(* Remove the empty constraints from a set if possible *)
(* be careful when to use it, it could cause problems if misused *)
(* Return true if g is a subset of some gg in css and hence ([], g) can be safely removed *)
let remove_empty_constraints_constr (g:var_env) (css:constr_set_set) : bool =
  let css' = List.filter (fun (_,gg) -> var_env_subset g gg) css in
  css' <> []

let remove_empty_constraints (css:constr_set_set) : constr_set_set =
    let rec rem_empty_aux (css_todo:constr_set_set) (css_done:constr_set_set) : constr_set_set =
    match css_todo with
    | [] -> css_done
    | ([],g) :: css_todo' ->
       let b = remove_empty_constraints_constr g (css_done@css_todo') in
       if b then rem_empty_aux css_todo' css_done
       else rem_empty_aux css_todo' (([],g)::css_done)
    | cs :: css_todo' -> rem_empty_aux css_todo' (cs::css_done)
    in
    rem_empty_aux css []
       
  
(* Typing rules for processes *)
let rec tcp (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (p:process) (q:process) : constr_set_set =
  (* This part takes care of rule TOr *)
  let top_or = collect_top_or g p q in 
  if top_or <> [] then
    let x = List.hd top_or in 
    let (t1,t2) = 
      match lookup_var g x with 
      | Some (TOr (t1,t2)) -> t1,t2 
      | _ -> failwith "This has to be an or type"
    in (tcp e e' (extend_var_env g x t1) d p q) @ (tcp e e' (extend_var_env g x t2) d p q)

  else
    match p, q with 
    (* PZero *)
    | PZero er , PZero er'-> 
        [([], g)] (* Branching is done in a final step *)

    (* POut *)
    | POut (er, u, p'), POut (er', v, q') ->
      let b,c = tct_new e e' g d u v (TLabel LL) in 
      if b then
        union_forall c (tcp e e' g d p' q') 
      else 
        raise (Type_Error ("Output message " ^ term2string u ^ " ~ " ^ term2string v ^ " cannot be typed with type L."(* ^ "Inferred type is " ^ type2string (fst (tct_real e e' g d u v))*),er,er'))

    (* PIn *)
    | PIn (er, x, p'), PIn (er', y, q') ->
      if x = y then
        if lookup_var g x = None then
          let css = tcp e e' (extend_var_env_nou g x (TLabel LL)) d p' q' in
          clean_constr_set_set_var css x
        else 
          raise (Type_Error ("Variable " ^ var2string x ^ " used in in was already bound",er,er'))
      else 
        raise (Type_Error ("Variables " ^ var2string x ^ " and " ^ var2string y ^ " used in in are not equal ",er,er'))

    (* PNew *)
    | PNew (er, n, l, p'), PNew (er',m,l',q') -> 
       if n = m && l = l' then
         if e n = Some l && e' n = Some l then 
           tcp e e' g d p' q'
         else 
           raise (Type_Error ("Annotated nonce type for " ^ nonce2string n ^ " does not match environment type",er,er'))
       else
         raise (Type_Error ("New nonce declaration does not match: " ^ nonce2string n ^ ":" ^ label2string l ^ " on the left, and " ^ nonce2string m ^ ":" ^ label2string l' ^ " on the right", er, er'))

    (* PNewKey *)
    | PNewKey (er, k, t, p'), PNewKey (er', k', t', q') -> 
       (match lookup_key d (k,k') with
       | Some t'' -> 
         if k = k' then
           if same_type t t' && same_type t t'' then
             tcp e e' g d p' q'
           else
             raise (Type_Error ("Annotated key type for " ^ k ^ " does not match environment type",er,er))
         else
           raise (Type_Error ("keys must be equal",er,er))
       | _ -> failwith ("Keytype should be inferred at start for " ^ k ^ " " ^ k'))

    (* PPar *)
    | PPar (er, p', p''), PPar (er', q', q'') ->
      add_cart (tcp e e' g d p' q') (tcp e e' g d p'' q'')

    | PRepl (er, p'), PRepl (er', q') -> failwith "Replication should not occure when typing."

    (* PLet + PLetLRK *)
    | PLet (er, x, de, p', p''), PLet (er', x', de', q', q'') ->
       if x = x' then 
         if lookup_var g x = None then 
           let css_ref, b_ref = 
             (match lookup_var g (dest_var de) with
              (*PLetLRK*)
              | Some (TRefine (VNonce (_,Nonce _),VNonce (_, Nonce _)))
              | Some (TRefine (VNonce (_,NonceStar _),VNonce (_, NonceStar _)))
              | Some (TRefine (VConst _,VConst _))
              | Some (TKey _)
              | Some (TEqKey _)
              | Some (TSessKey _) ->
                 tcp e e' g d p'' q'', true
              | Some _ -> [], false
              (* PLet *)
              | None -> raise (Type_Error ("Variable " ^ var2string (dest_var de) ^ " used in let is not bound",er,er')))
           in
           let b_ref = b_ref && same_de de de' in 
           (* Tying with the LRK/incompatible rule was successful *)
           if b_ref then css_ref else
             (*otherwise PLetDec/Adec*)
             let ench_css, ench_b = 
               (match lookup_var g (dest_var de) with 
                (* These two cases should be handled by the original destructor rules *)
                | Some (TSEnc _) -> [], false
                | Some (TAEnc _) -> [], false
                | Some (TLabel LL) -> 
                   (match de, de' with 
                    | DSDec (er, y, VKey (_, k)), DSDec (er', y', VKey (_, k')) -> 
                       if y <> y' then raise (Type_Error ("Variables must be equal",er, er')) else
                         (match lookup_key d (k,k') with 
                          | Some TSessKey (LS, t) ->
                             let css1 = tcp e e' (extend_var_env_nou g x t) d p' q' in
                             let css1' = clean_constr_set_set_var css1 x in
                             let css2 = tcp e e' g d p'' q'' in
                             remove_empty_constraints (List.sort_uniq Pervasives.compare (css2 @ css1')), true
                          | Some TEqKey (LS, t)
                          | Some TKey (LS, t) -> 
                             let css1 = tcp e e' (extend_var_env_nou g x t) d p' q' in
                             let css1' = clean_constr_set_set_var css1 x in
                             let css2 = tcp e e' g d p'' q'' in 
                             let l1,l2 = key_combinations d k k' in 
                             let css3 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p' q'') [] l1 in
                             let css3' = clean_constr_set_set_var css3 x in
                             let css4 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p'' q') [] l2 in 
                             let css4' = clean_constr_set_set_var css4 x in
                             remove_empty_constraints (List.sort_uniq Pervasives.compare (css2 @ css1' @ css3' @ css4')), true
                          | _ ->  [], false)
                    | DADec (er, y, VKey (_, k)), DADec (er', y', VKey (_, k')) -> 
                       if y <> y' then raise (Type_Error ("Variables must be equal",er, er')) else
                         (match lookup_key d (k,k') with 
                          | Some TSessKey (LS, t) ->
                             let css1 = tcp e e' (extend_var_env_nou g x t) d p' q' in
                             let css1' = clean_constr_set_set_var css1 x in
                             let css2 = tcp e e' g d p'' q'' in
                             remove_empty_constraints (List.sort_uniq Pervasives.compare (css2 @ css1')), true
                          | Some TEqKey (LS, t)
                          | Some TKey (LS, t) -> 
                             if k = k' then (*PLetAdecSame*)
                               let css1 = tcp e e' (extend_var_env_nou g x t) d p' q' in
                               let css1' = clean_constr_set_set_var css1 x in
                               let css2 = tcp e e' (extend_var_env_nou g x (TLabel LL)) d p' q' in 
                               let css2' = clean_constr_set_set_var css2 x in 
                               let css3 = tcp e e' g d p'' q'' in 
                               let l1,l2 = key_combinations d k k' in 
                               let css4 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p' q'') [] l1 in
                               let css4' = clean_constr_set_set_var css4 x in
                               let css5 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p'' q') [] l2 in 
                               let css5' = clean_constr_set_set_var css5 x in
                               remove_empty_constraints (List.sort_uniq Pervasives.compare (css1' @ css2' @ css3 @ css4' @ css5')), true
                             else (*PLetAdecDiff*)
                               let css1 = tcp e e' (extend_var_env_nou g x t) d p' q' in
                               let css1' = clean_constr_set_set_var css1 x in
                               let css2 = tcp e e' (extend_var_env_nou g x (TLabel LL)) d p' q'' in 
                               let css2' = clean_constr_set_set_var css2 x in 
                               let css3 = tcp e e' (extend_var_env_nou g x (TLabel LL)) d p'' q' in 
                               let css3' = clean_constr_set_set_var css3 x in 
                               let css4 = tcp e e' g d p'' q'' in 
                               let l1,l2 = key_combinations d k k' in 
                               let css5 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p' q'') [] l1 in 
                               let css5' = clean_constr_set_set_var css5 x in 
                               let css6 = List.fold_left (fun css t' -> css @ (tcp e e' (extend_var_env_nou g x t')) d p'' q') [] l2 in 
                               let css6' = clean_constr_set_set_var css6 x in 
                               remove_empty_constraints (List.sort_uniq Pervasives.compare ( css1' @ css2'@css3'@css4@css5'@css6')),true
                          | _ -> [], false)
                    | _ -> [], false )
                | _ -> [], false) in 
             
             (* Typing with PLetDec/Adec rules was successful *)
             if ench_b then ench_css else
               (*otherwise: PLet*)
               if same_de de de' then
                 (match lookup_var g (dest_var de) with
                    | Some tv -> 
                       let t = tcd e e' g d de de' in                   
                       let css1 = tcp e e' (extend_var_env_nou g x t)  d p' q' in
                       let css1' = clean_constr_set_set_var css1 x in 
                       let css2 = tcp e e' g d p'' q'' in
                       remove_empty_constraints (List.sort_uniq Pervasives.compare (css2 @ css1'))
                    | None -> failwith "Impossible case in EncH")
               else
                 raise (Type_Error ("Structure of the two let expressions is different",er,er'))
         else
           raise (Type_Error ("Variable "  ^ var2string x ^ " used in let is already bound",er,er'))
       else
         raise (Type_Error ("The two variables are not equal",er,er'))

    (* PIf *)
    | PIf (er, u, v, p', p''), PIf (er', u', v', q', q'') ->
       let aa =
         (match infer_refinement e e' g d u u', infer_refinement e e' g d v v' with
          | Some (TRefine (u1,u2)), Some (TRefine (v1,v2)) ->
             ((*print_endline ((term2string u)^"~"^(term2string u')^" : "^(type2string (TRefine (u1,u2))));
             print_endline ((term2string v)^"~"^(term2string v')^" : "^(type2string (TRefine (v1,v2))));*)
             (* PIfLR* *)
             if is_star_nonce u1 && is_star_nonce u2 && is_star_nonce v1 && is_star_nonce v2 then
               if same_term u1 v1 && same_term u2 v2 then
                 Some (tcp e e' g d p' q' @ tcp e e' g d p'' q'')
               else
                 None
             else
               (* PIfLR'* *)
               if not (is_star_nonce u1) && not (is_star_nonce u2) && is_star_nonce v1 && is_star_nonce v2 then
                 Some (tcp e e' g d p' q' @ tcp e e' g d p'' q'') 
                                          (* symmetric variant of PIfLR'* *)
               else
              if not (is_star_nonce v1) && not (is_star_nonce v2) && is_star_nonce u1 && is_star_nonce u2 then
                Some (tcp e e' g d p' q' @ tcp e e' g d p'' q'')
              (* PIfLR *)
              else
                if not (is_star_nonce u1) && not (is_star_nonce u2) && not (is_star_nonce v1) && not (is_star_nonce v2) then
                (match same_term u1 v1, same_term u2 v2 with 
                 | true, true -> Some (tcp e e' g d p' q')
                  | true, false -> Some (tcp e e' g d p' q'')
                  | false, true -> Some (tcp e e' g d p'' q')
                  | false, false -> Some (tcp e e' g d p'' q''))
                else
                  None)
          | _ -> None) in
       (match aa with
       | Some aaa -> aaa
       | None -> 

          (* PIfS +_PIf + PIf* *) 
          if incompatible_types e e' g d u u' v v'  then
            tcp e e' g d p'' q'' 
          else
            (* PIfL *)
            let bt1, ct1 = tct_new e e' g d u u' (TLabel LL) in
            let bt2, ct2 = tct_new e e' g d v v' (TLabel LL) in
            if bt1 && bt2 then 
              let cp1 = tcp e e' g d p' q' in 
              let cp2 = tcp e e' g d p'' q'' in 
              (* PIfP *)
              let pifp = 
                (match v, v' with 
                 | VKey (_,t), VKey (_,t') -> t = t'
                 | VConst (_,t), VConst (_,t') -> t = t'
                 | VNonce (_,t), VNonce (_,t') -> t = t'
                 | _ -> false) in 
              if pifp then
                match p'',q'' with
                | PZero _, PZero _ -> cp1
                | _, _ -> cp1 @ cp2
              else
                match p'',q'' with
                | PZero _, PZero _ -> union_forall (ct1 @ ct2) cp1
                | _, _ -> union_forall (ct1 @ ct2) (cp1 @ cp2)
            else
              ((* TODO: new rule to support private authentication *)
              print_verbose 2 ("New If RULE");
              let cp1 = tcp e e' g d p' q' in 
              let cp2 = tcp e e' g d p' q'' in 
              let cp3 = tcp e e' g d p'' q' in 
              let cp4 = tcp e e' g d p'' q'' in 
              cp1 @ cp2 @ cp3 @ cp4))     
    | _ -> raise (Type_Error  ("Case not covered by typing rules:\n" ^ process2string p ^ "\n" ^ process2string q,process_error p ,process_error q))
     


let collectnv_cs ((cs):constr_set) : nonce list * var list = 
  List.fold_left 
  (fun (nl,vl) c -> 
    let ns,vs = collectnv_c c in 
    List.sort_uniq Pervasives.compare (nl @ ns), 
    List.sort_uniq Pervasives.compare (vl @ vs)) 
  ([],[]) cs

(* We implement branching only for the variables that appear in the
 * constraints, not for all *)
let branching (cs:constr_set_env) : constr_set_set = let g = snd cs in 
  let v_or =  
    List.filter (fun x -> match lookup_var g x with Some (TOr (t1, t2)) -> true | _ -> false) 
      (snd (collectnv_cs (fst cs))) in 
  List.fold_left 
    (fun css x -> match lookup_var g x with 
      | Some (TOr (t1,t2)) -> 
        List.map (fun (cs,g') -> (cs, extend_var_env g' x t1)) css  @ 
        List.map (fun (cs,g') -> (cs, extend_var_env g' x t2)) css
      | _ -> failwith "List should only contain Or types"
      )
    [cs] v_or


let branching_css (css:constr_set_set) : constr_set_set = 
  List.flatten (List.map branching css) 

