open Syntax

let __verbose_level = ref 0
(* Utility Functions *)

(* 2string functions *)

let nonce2string (n:nonce) : string = 
  match n with | Nonce s | NonceStar s -> s

let var2string (x:var) : string = x

let rec term2string (t:term) : string  = match t with 
  | VNonce (_, n) -> nonce2string n
  | VConst (_, c) -> c
  | VVar (_, x) -> var2string x
  | VKey (_, k) -> k 
  | VPKey (_, t) -> "pk("^ term2string t ^")"
  | VVerKey (_, t) -> "vk("^ term2string t ^")"
  | VPair (_, u, v) -> "("^ term2string u ^","^term2string v^")"
  | VInl (_, v) -> "inl("^term2string v^")"
  | VInr (_, v) -> "inr("^term2string v^")"
  | VHash (_, v) -> "h("^term2string v^")"
  | VSEnc (_, v, t) -> "enc("^term2string v^","^ term2string t ^")"
  | VAEnc (_, v, t) -> "aenc("^term2string v^","^ term2string t ^")"
  | VSign (_, v, t) -> "sign("^term2string v^","^ term2string t ^")"

let label2string (l:label) : string = match l with 
| LL -> "L"
| LH -> "H"
| LS -> "S"

let rec type2string (t:ty) : string = match t with
| TLabel l -> label2string l
| TPair (t1, t2) -> "(" ^ type2string t1 ^" * "^type2string t2 ^ ")"
| TOr (t1, t2) -> "(" ^ type2string t1 ^" v "^type2string t2 ^ ")"
| TSum (t1, t2) -> "(" ^ type2string t1 ^" + "^type2string t2 ^ ")"
| TSessKey (l, t) -> "seskey("^label2string l^", "^type2string t ^ ")"
| TEqKey (l, t) -> "eqkey("^label2string l^", "^type2string t ^ ")"
| TPubKey t -> "pkey("^type2string t ^ ")"
| TVerKey t -> "vkey("^type2string t ^ ")"
| TKey (l, t) -> "key("^label2string l^", "^type2string t ^ ")"
| TRefine (u,v) -> "{x_L=" ^ term2string u ^ ", x_R="^term2string v^"}"
| TSEnc (t,t') -> "(" ^ type2string t ^ ")_" ^ type2string t' 
| TAEnc (t,t') -> "{" ^ type2string t ^ "}_" ^ type2string t' 

let dest2string (d:dest) : string = match d with
| DSDec (_,x, t) -> "dec("^var2string x^","^ term2string t ^")"
| DADec (_,x, t) -> "adec("^var2string x^","^ term2string t ^")"
| DProj1 (_,p) -> "proj_1("^var2string p^")"
| DProj2 (_,p) -> "proj_2("^var2string p^")"
| DVer (_,x, t) -> "ver("^var2string x^", " ^ term2string t ^ ")"

let rec process2string (p:process) : string = match p with
| PZero e  -> "0"
| PNew (e, n, l, q) -> "new " ^ nonce2string n ^ ":" ^ label2string l ^ "." ^ process2string q
| PNewKey (e, k, t, q) -> "new " ^ k ^ ":" ^ type2string t ^ "." ^ process2string q
| POut (e, t, q) -> "out(" ^ term2string t ^ ")." ^ process2string q
| PIn (e, x, q) -> "in(" ^ var2string x ^ ")." ^ process2string q
| PPar (e, q1, q2) -> "(" ^ process2string q1 ^ " | " ^ process2string q2 ^ ")"
| PRepl (e, p') -> "!(" ^ process2string p' ^ ")"
| PLet (e, x, d, q1, q2) -> "let " ^ var2string x ^ "=" ^ dest2string d ^ " in " ^ process2string q1 ^ " else " ^ process2string q2
| PIf (e, u, v, q1, q2) -> "if " ^ term2string u ^ "=" ^ term2string v ^ " then " ^ process2string q1 ^ " else " ^ process2string q2
| PMatch (e, x, y, q1, z, q2) -> "match " ^ var2string x ^ " with | inl(" ^ var2string y ^ ") -> " ^ process2string q1 ^ " | inr(" ^ var2string z ^ ") -> " ^ process2string q2 

let constr2string ((u,v):constr) : string = term2string u ^" ~ "^ term2string v
let constrset2string (cs:constr_set) : string =  
  "{" ^ String.concat "\n" (List.map constr2string cs) ^ "}" 
let gamma2string (g:var_env): string =
  "[" ^
    String.concat ", "
                  (VarEnv.fold (fun x -> fun t -> fun l ->  ((var2string x ^ ":" ^ type2string t)::l)) g [])
  ^ "]"
let constrsetenv2string ((cs, g):constr_set_env) : string =  
  constrset2string cs ^ "  " ^ gamma2string g 
let constrsetset2string (css:constr_set_set) : string =  
  "{" ^ String.concat "\n\n\n" (List.map constrsetenv2string css) ^ "}"

let print_verbose (v:int) (s:string) : unit =
  if !__verbose_level >= v then print_endline s
 
  
(* equality check funcitons (necessary becasue defaultt equality is broken by
 * different error positions) *)

let rec same_term (t:term) (t':term) : bool = 
  match t,t' with 
  | VNonce (_, n), VNonce (_,n') -> n = n'
  | VConst (_, c), VConst (_,c') -> c = c'
  | VVar (_, x), VVar (_,x') -> x = x'
  | VKey (_, k), VKey (_,k') -> k = k'
  | VPKey (_, k), VPKey (_,k') -> same_term k k'
  | VVerKey (_, k), VVerKey (_,k') -> same_term k k'
  | VPair (_, u, v), VPair (_,u',v') -> same_term u u' && same_term v v'
  | VInl (_, v), VInl (_,v') -> same_term v v'
  | VInr (_, v), VInr (_,v') -> same_term v v'
  | VHash (_, v), VHash (_,v') -> same_term v v'
  | VSEnc (_, v, k), VSEnc (_,v',k') -> same_term k k' && same_term v v'
  | VAEnc (_, v, k), VAEnc (_,v',k') -> same_term k k' && same_term v v'
  | VSign (_, v, k), VSign (_,v',k') -> same_term k k' && same_term v v'
  | _ -> false


let rec same_type (t:ty) (t':ty) : bool = 
  t = t' || 
  match t,t' with 
  | TLabel l, TLabel l'  -> l = l' 
  | TPair (t1, t2), TPair (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TOr (t1, t2), TOr (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TSum (t1, t2), TSum (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TKey (l, t), TKey (l',t') -> l = l' && same_type t t'
  | TRefine (u,v), TRefine (u',v') -> same_term u u' && same_term v v'
  | TSEnc (t1,t2), TSEnc (t1', t2') -> same_type t1 t1' && same_type t2 t2'
  | TAEnc (t1,t2), TAEnc (t1', t2') -> same_type t1 t1' && same_type t2 t2'
  | TPubKey t, TPubKey t' -> same_type t t'
  | TVerKey t, TVerKey t' -> same_type t t'
  | TSessKey (l, t), TSessKey (l',t') -> l = l' && same_type t t'
  | TEqKey (l, t), TEqKey (l',t') -> l = l' && same_type t t'
  | _ -> false

let same_de (de:dest) (de':dest) : bool = 
match de,de' with
| DSDec (_,x, t), DSDec (_,x', t') -> x = x' && same_term t t'
| DADec (_,x, t), DADec (_,x', t') -> x = x' && same_term t t'
| DProj1 (_,p), DProj1 (_,p')-> p = p'
| DProj2 (_,p), DProj2 (_,p')-> p = p'
| DVer (_,x, t), DVer (_,x', t') -> x = x' && same_term t t'
| _ -> false

(* Commonly used access to subcomponents *)

let dest_var (de:dest) : var = match de with
| DSDec (_,x, _) -> x
| DADec (_,x, _) -> x
| DProj1 (_,p) -> p
| DProj2 (_,p) -> p
| DVer (_,x, _) -> x

let dest_err (de:dest) : error = match de with
| DSDec (e,x, _) -> e
| DADec (e,x, _) -> e
| DProj1 (e,p) -> e
| DProj2 (e,p) -> e
| DVer (e,x, _) -> e

let term_error (t:term) : error = match t with
| VNonce (er, _) 
| VConst (er, _) 
| VVar (er, _) 
| VKey (er, _) 
| VPKey (er, _) 
| VVerKey (er, _) 
| VPair (er,_, _) 
| VInl (er, _)
| VInr (er, _)
| VHash (er, _) 
| VSEnc (er, _, _) 
| VAEnc (er, _, _)
| VSign (er, _, _) -> er

let process_error (p:process) : error = match p with
| PZero e  -> e
| PNew (e, n, l, q) -> e
| PNewKey (e, k, t, q) -> e
| POut (e, t, q) -> e
| PIn (e, x, q) -> e
| PPar (e, q1, q2) -> e
| PRepl (e, p') -> e
| PLet (e, x, d, q1, q2) -> e
| PIf (e, u, v, q1, q2) -> e
| PMatch (e, x, y, q1, z, q2) -> e

let is_nonce (t:term) : bool= 
  match t with 
  | VNonce _ -> true
  | _ -> false

let is_star_nonce (t:term) : bool = 
  match t with 
  | VNonce (_, (NonceStar _)) -> true
  | _ -> false

let rec no_star (t:term) : bool = 
  match t with
  | VNonce (_, n) -> not (is_star_nonce t)
  | VConst (_, c) -> true
  | VVar (_, x) -> true
  | VKey (_, k) -> true 
  | VPKey (_, t') -> no_star t'
  | VVerKey (_, t') -> no_star t'
  | VPair (_, u, v) -> no_star u && no_star v
  | VInl (_, v) -> no_star v
  | VInr (_, v) -> no_star v
  | VHash (_, v) -> no_star v
  | VSEnc (_, v, t') -> no_star v && no_star t'
  | VAEnc (_, v, t') -> no_star v && no_star t'
  | VSign (_, v, t') -> no_star v && no_star t'


(* Utility functions *)

let rec subtyping (t1:ty) (t2:ty) = 
  (* SRefl *)
  same_type t1 t2 ||
  match t1, t2 with 
  (* SHigh *)
  | _ , TLabel LH -> true
  (* SPair *)
  | TPair (t1', t1''), TPair (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SPairL *)
  | TPair (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SPairS  + SPairS'*)
  | TPair (t1', t1''), TLabel LS -> subtyping t1' (TLabel LS) ||  subtyping t1'' (TLabel LS)
  (* SSum *)
  | TSum (t1', t1''), TSum (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SSumL *)
  | TSum (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SKey *)
  | TKey (l,t), TLabel l' -> l = l'
  (* SEnc *)
  | TSEnc (t1,t2), TSEnc (t1',t2') -> subtyping t1 t1' && same_type t2 t2'
  (* SAEnc *)
  | TAEnc (t1,t2), TAEnc (t1',t2') -> subtyping t1 t1' && same_type t2 t2'
  (* Rule STrans is incomporated directly in the other rules *)
  | _ -> false
   



let env_empty (x:'a) : 'b option = None
let var_env_empty () : var_env = VarEnv.empty

let extend (env: 'a -> 'b option) (k:'a) (v:'b) (x:'a) : 'b option = if x = k then Some v else env x
let extend_nou (env: 'a -> 'b option) (k:'a) (v:'b) (x:'a) : 'b option = if env k = None then if x = k then Some v else env x else failwith "OVERWRITE"

let lookup_var (g:var_env) (x:var) : ty  option =
  try 
    let t = (VarEnv.find x g) in
    print_verbose 3 ("Looking up " ^ x ^ " : " ^ type2string t);
    Some t
  with
  | _ -> None 

let extend_var_env (g:var_env) (x:var) (t:ty) : var_env =
  print_verbose 2 ("Variable " ^ x ^ " has type " ^ type2string t); 
  VarEnv.add x t g

let extend_var_env_nou (g:var_env) (x:var) (t:ty) : var_env =
  match lookup_var g x with
  | Some _ -> failwith ("Environment may not be updated at this point " ^ x ^ ": " ^ type2string t) 
  | None -> extend_var_env g x t


exception UnionError 

let var_env_union (g:var_env) (g': var_env) : var_env option = 
  try
    (*we could use the ocaml 4.03 union function, but for compatibility we use merge*)
    Some (VarEnv.merge
            (fun x ot1 ot2 ->
              match ot1,ot1 with
              | Some t1, Some t2 ->
                 if same_type t1 t2 then Some t1 
                 else raise UnionError
              | None, _ -> ot2
              | _, None -> ot1) g g')
  with
  | UnionError -> None


let key_is_repl (d:key_env) (k:key) : bool =
  try 
    let t = snd (KeyEnv.find (k,k) d) in 
    print_verbose 3 ("Looking up repl of  " ^ k ^ " : " ^ string_of_bool t);
    t
  with
  | _ -> false 

let lookup_key (g:key_env) (x:key * key) : ty  option =
  try 
    let t = fst (KeyEnv.find x g) in 
    print_verbose 3 ("Looking up " ^ fst x ^ "," ^ snd x ^ " : " ^ type2string t);
    Some t
  with
  | _ -> None 

let extend_key_env (d:key_env) (x:key*key) (t:ty) (repl:bool) : key_env =
  KeyEnv.add x (t,repl) d


let get_key_label (g:key_env) (k:key) : label option =
  match lookup_key g (k,k) with 
  | Some (TSessKey (l, _)) 
  | Some (TEqKey (l, _)) 
  | Some (TKey (l, _)) -> Some l
  | None -> 
      let t = (
        try Some (snd (List.find (fun ((k1,k2),t) -> k1 = k || k2 = k) (KeyEnv.bindings g))) with 
        | Not_found -> None)  in
      (match t with
      | Some (TSessKey (l, _),_) 
      | Some (TEqKey (l, _),_) 
      | Some (TKey (l, _),_) -> Some l
      | None -> None
      | _ -> failwith "Key environment may only contain key types")
  | _ -> failwith "Key environment may only contain key types"

let is_in_keys (g:key_env) (k:key) : bool = 
  match get_key_label g k with 
  | Some _ -> true
  | None -> false

let key_combinations (g:key_env) (k1:key) (k2:key) : (ty list * ty list) = 
  let l = KeyEnv.bindings g in 
  let extract_types ll = List.sort_uniq Pervasives.compare (List.map (fun x -> match fst (snd x) with | TKey (_ ,t') -> t' | _ -> failwith "Key environment can contain only key types") ll) in
  let l1 = extract_types (List.filter (fun ((k1',k2'), t) -> k1' = k1 && k2' <> k2) l) in 
  let l2 = extract_types (List.filter (fun ((k1',k2'), t) -> k2' = k2 && k1' <> k1) l) in 
  l1, l2


let add_forall (c:constr) (css:constr_set_set) : constr_set_set =  
  List.map (fun (cs,g) -> (c::cs,g)) css
let union_forall (cs:constr_set) (css:constr_set_set) : constr_set_set = 
  List.map (fun (cs',g) -> (cs@cs',g)) css
let add_cart (css:constr_set_set) (css':constr_set_set) : constr_set_set = 
  List.flatten 
  (List.map (fun (cs, g) -> List.map (fun (cs', g') -> match var_env_union g g' with | Some gn -> cs@ cs', gn | None -> [], var_env_empty ()) css') css)


(* instantiate all variables with refinement types *)
let rec instantiate (g:var_env) (t:term) (left:bool) (star:bool): term = 
  let res = 
    match t with
    | VNonce (_,n) -> t
    | VConst (_,c) -> t
    | VVar (_,x) -> 
        (match lookup_var g x with 
        | Some (TRefine (t1,t2)) -> 
          if star || (no_star t1 && no_star t2) then
            if left then t1 else t2
          else
            t
        | _ -> t)
    | VKey (_,k) -> t
    | VPKey _ -> t
    | VVerKey _ -> t
    | VPair (er,t1, t2) -> 
        let t1' = instantiate g t1 left star in 
        let t2' = instantiate g t2 left star in 
        VPair (er,t1', t2')
    | VInl (er,v) -> VInl  (er,instantiate g v left star)
    | VInr (er,v) -> VInr (er,instantiate g v left star)
    | VHash (er,v) -> VHash (er,instantiate g v left star)
    | VSEnc (er, v, k) -> VSEnc (er,(instantiate g v left star), k)
    | VAEnc (er, v, k) -> VAEnc (er,(instantiate g v left star), k)
    | VSign (er, v, k) -> VSign (er,(instantiate g v left star), k)
  in 
(*
  print_verbose 2 "Instantiate :" ;
  print_verbose 2 (term2string t ^ " to " ^ term2string res ^ "\n") ; 
*)
  res

(* Check if term contains a nonce or key with label S directly under a pair *)
let rec s_under_pair (e:nonce_env) (g:var_env) (d:key_env) (t:term) : bool = 
  match t with 
  | VNonce (_, n) -> 
      (match e n with 
       | Some LS -> true 
       | _ -> false)
  | VKey (_, k) -> 
      (match get_key_label d k with 
      | Some LS -> true 
      | _ -> false)
  | VVar (_, x) ->
     (match lookup_var g x with
      (*    (*in principle we should be able to do this but we cannot since we do not keep the label in the refinement*)
      | Some (TRefine (VNonce (_, n), VNonce (_, m))) ->
         (match e n, e m with
          | Some LS, Some LS -> true
          | _, _ -> false)*)
      | Some (TKey (LS, _)) -> true
      | Some (TEqKey (LS, _)) -> true
      | Some (TSessKey (LS, _)) -> true
      | _ -> false)
  | VPair (_,t1,t2) -> s_under_pair e g d t1 || s_under_pair e g d t2
  | _ -> false

(* Check if a term contains a star-subterm *)


let rec contains_star_t (t:term) : bool  = match t with 
  | VNonce (_, n) -> 
      (match n with 
      | Nonce _ -> false
      | NonceStar _ -> true)
  | VConst (_, c) -> false
  | VVar (_, x) -> false
  | VKey (_, k) -> false
  | VPKey (_, t') -> contains_star_t t'
  | VVerKey (_, t') -> contains_star_t t'
  | VPair (_, u, v) -> contains_star_t u || contains_star_t v 
  | VInl (_, v) -> contains_star_t v
  | VInr (_, v) -> contains_star_t v
  | VHash (_, v) -> contains_star_t v
  | VSEnc (_, v, t') -> contains_star_t v || contains_star_t t'
  | VAEnc (_, v, t') -> contains_star_t v || contains_star_t t' 
  | VSign (_, v, t') -> contains_star_t v || contains_star_t t' 

(* collect all, variables and keys of a term *)
let rec collectnv_t (t:term) : (nonce list * key list *  var list) = 
  match t with
  | VNonce (_, n) -> ([n], [], [])
  | VConst (_, c) -> ([], [], [])
  | VVar (_, x) -> ([], [], [x]) 
  | VKey (_, k) -> ([], [k], []) 
  | VPKey (_, t') -> collectnv_t t'
  | VVerKey (_, t') -> collectnv_t t'
  | VInl (_,v) -> collectnv_t v
  | VInr (_,v) -> collectnv_t v
  | VHash (_,v) -> collectnv_t v
  | VSEnc (_, u, v)
  | VAEnc (_, u, v)
  | VSign (_, u, v)
  | VPair (_,u, v) -> 
      let (n1,k1,v1) = collectnv_t u in 
      let (n2,k2,v2) = collectnv_t v in 
      (List.sort_uniq Pervasives.compare (n1 @ n2),
      List.sort_uniq Pervasives.compare (k1 @ k2),
      List.sort_uniq Pervasives.compare (v1 @ v2))

let ground (t:term) : bool = let (_, _, v) = collectnv_t t in v = []

(* collect nonces and variables from constraints *)
let collectnv_c ((u,v):constr) : nonce list * var list  = 
  let (n_u, _, v_u) = collectnv_t u in 
  let (n_v, _, v_v) = collectnv_t v in 
  List.sort_uniq Pervasives.compare (n_u @ n_v),
  List.sort_uniq Pervasives.compare (v_u @ v_v)

let well_formed_refinement (t:term) (t':term) : bool  = 
  ground t && ground t' && (is_star_nonce t && is_star_nonce t') || (no_star t && no_star t')

let rec concat_news_process (news:process) (q:process) : process =
  match news with 
  | PZero _ -> q
  | PNew (er,n,l,p) -> PNew (er,n,l, concat_news_process p q)
  | PNewKey (er,k,t,p) -> PNewKey (er,k,t, concat_news_process p q)
  | _ -> failwith "Process may only contain 'new' instructions"


(* unused *)
(* remove from g the bindings for variables which do not appear in cs *)
(*let trim_gamma (cs:constr_set) (g:var_env) : var_env =
  let vs = List.flatten (List.map (fun c -> snd (collectnv_c c)) cs) in
  VarEnv.filter (fun x -> fun t -> List.mem x vs) g*)
       
(* remove from g the binding for x if x does not appear in cs, and do nothing otherwise *)
let trim_gamma_var (cs:constr_set) (g:var_env) (x:var) : var_env =
  let vs = List.flatten (List.map (fun c -> snd (collectnv_c c)) cs) in
  if List.mem x vs then
    g
  else
    VarEnv.filter (fun y -> fun t -> x <> y) g

let remove_dup_constr_set (cs:constr_set) : constr_set =
  List.sort_uniq Pervasives.compare cs

let clean_constr_set_env_var ((cs,g):constr_set_env) (x:var) : constr_set_env =
  let cs'= remove_dup_constr_set cs in
  (cs',trim_gamma_var cs' g x)

let clean_constr_set_set_var css x = 
  let css' = List.map (fun c -> clean_constr_set_env_var c x) css in
  let css'' = List.sort_uniq Pervasives.compare css' in
  let l = (List.length css) - (List.length css'') in
  print_verbose 2 ("--> simplifying constraints: " ^ string_of_int l^" constraint(s) removed (variable: " ^ (var2string x)^")");
  css''

type side = Left | Right

(* could use a more efficient representation *)
type substitution = (var * term) list

let subs_var (s:substitution) (x:var) : term option =
  try
    Some (List.assoc x s)
  with
    Not_found -> None
               
                  
(* was previously named instantiate *)
let rec substitute (s:substitution) (u:term) : term = 
  match u with
  | VNonce (_,n) -> u
  | VConst (_,c) -> u
  | VVar (_,x) -> 
      (match subs_var s x with 
      | Some v -> v
      | None -> u)
  | VKey _ -> u
  | VPKey (er,t') -> VPKey (er, substitute s t')
  | VVerKey (er,t') -> VVerKey (er, substitute s t')
  | VPair (er, u1, u2) -> 
      let u1' = substitute s u1 in 
      let u2' = substitute s u2 in 
      VPair (er, u1', u2')
  | VInl (er,v) -> VInl  (er, substitute s v)
  | VInr (er,v) -> VInr (er, substitute s v)
  | VHash (er,v) -> VHash (er, substitute s v)
  | VSEnc (er, u, v) -> VSEnc (er, substitute s u, substitute s v)
  | VAEnc (er, u, v) -> VAEnc (er, substitute s u, substitute s v)
  | VSign (er, u, v) -> VSign (er, substitute s u, substitute s v)


let option_get (x: 'a option) : 'a = 
  match x with 
  | Some y -> y
  | None -> failwith "Expected Some"
